@@ -1,10 +1,10 @@
 /* Conversion of files between different charsets and surfaces.
-   Copyright  1990,92,93,94,96,97,98,99,00 Free Software Foundation, Inc.
-   Franois Pinard <pinard@iro.umontreal.ca>, 1990.
+   Copyright © 1990,92,93,94,96,97,98,99,00 Free Software Foundation, Inc.
+   François Pinard <pinard@iro.umontreal.ca>, 1990.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
-   Free Software Foundation; either version 2, or (at your option) any later
+   Free Software Foundation; either version 3, or (at your option) any later
    version.
 
    This program is distributed in the hope that it will be useful, but
@@ -18,37 +18,19 @@
 
 #include "common.h"
 
+#include <stdlib.h>
+#include <stdio.h>
 #include <ctype.h>
-
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <unistd.h>
+#include <libgen.h>
+#include <utime.h>
+#include <setjmp.h>
+#include <locale.h>
 
 #include "getopt.h"
-
-#if HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-
-#if HAVE_STRUCT_UTIMBUF
-# if HAVE_UTIME_H
-#  include <utime.h>
-# endif
-#else
-
-struct utimbuf
-{
-  time_t actime;
-  time_t modtime;
-};
-
-#endif
-
-#include <setjmp.h>
-
-#if OS2
-# include <io.h>
-# include <fcntl.h>
-#endif
+#include "xbinary-io.h"
 
 /* Variables.  */
 
@@ -64,9 +46,6 @@ static int show_version = 0;
 /* If true, show a list of one or all known charsets, then exit.  */
 static bool show_symbols = false;
 
-/* If true, dump all tables as a big C module.  */
-static bool freeze_tables = false;
-
 /* If true, check all charsets for subsets, then exit.  */
 static bool find_subsets = false;
 
@@ -92,7 +71,7 @@ static bool quiet_flag = false;
 
 /* If the recoding yields some problems in reversibility in some file, this
    file replacement is denied and it is left unrecoded or, if recode is used
-   as a mere filter, the recoding is interrupted.  The following flag forces
+   as a mere filter, the recoding is aborted.  The following flag forces
    the recoding to run to completion and the replacement to be done even if
    the recoding is not reversible.  */
 static bool force_flag = false;
@@ -102,27 +81,16 @@ static bool force_flag = false;
    recoding the most transparent possible to make, and other tools.
    However, selecting the following option inhibit the timestamps handling,
    thus effectively `touching' the file.  */
-bool touch_option = false;
+static bool touch_option = false;
 
 /* With strict mapping, all reversibility fallbacks get defeated.  */
-bool strict_mapping = false;
+static bool strict_mapping = false;
 
 /* The following charset name will be ignored, if given.  */
 static const char *ignored_name = NULL;
 
-#if 0
-/* Unabridged names of BEFORE and AFTER charsets, even if still aliases.
-   These are used for naming the array in produced C code.  */
-static const char *before_full_name;
-static const char *after_full_name;
-#endif
-
 /* Ordinals of list, BEFORE and AFTER charset.  */
 static RECODE_SYMBOL list_charset;
-#if 0
-static RECODE_SYMBOL before_charset;
-static RECODE_SYMBOL after_charset;
-#endif
 
 /* Flag telling usage that we are decoding charsets.  */
 bool decoding_charset_flag = false;
@@ -168,33 +136,30 @@ task_perror (RECODE_CONST_TASK task)
 | Signal handler.  |
 `-----------------*/
 
-#if FIXME
-static jmp_buf signal_label;	/* where to jump when signal received */
-#endif
-static bool interrupted = 0;	/* set when some signal has been received */
-
-static RETSIGTYPE
-signal_handler (int number)
+#ifdef HAVE_PIPE
+static void
+sig_catch(int sig, void (*handler) (int))
 {
-  interrupted = 1;
-  signal (number, signal_handler);
+  struct sigaction sa;
+  sa.sa_handler = handler;
+  sa.sa_flags = 0;
+  sigemptyset (&sa.sa_mask);
+  sigaction (sig, &sa, NULL);  /* ignore error: none possible */
 }
 
-/*------------------------------------------------------------------------.
-| Prepare to handle signals, intercept willingful requests for stopping.  |
-`------------------------------------------------------------------------*/
+static void
+signal_handler (int number)
+{
+  recode_interrupted = 1;
+  sig_catch (number, signal_handler);
+}
 
 static void
 setup_signals (void)
 {
-#ifdef SIGPIPE
-  signal (SIGPIPE, signal_handler);
-#endif
-#if 0
-  signal (SIGINT, signal_handler);
-  signal (SIGTERM, signal_handler);
-#endif
+ sig_catch (SIGPIPE, signal_handler);
 }
+#endif
 
 /* Main control.  */
 
@@ -208,7 +173,7 @@ print_copyright (void)
   fputs (_("\
 This program is free software; you can redistribute it and/or modify\n\
 it under the terms of the GNU General Public License as published by\n\
-the Free Software Foundation; either version 2, or (at your option)\n\
+the Free Software Foundation; either version 3, or (at your option)\n\
 any later version.\n\
 \n\
 This program is distributed in the hope that it will be useful,\n\
@@ -235,7 +200,7 @@ usage (int status, bool list)
   else
     {
       fputs (_("\
-Free `recode' converts files between various character sets and surfaces.\n\
+Recode converts files between various character sets and surfaces.\n\
 "),
 	     stdout);
       printf (_("\
@@ -253,7 +218,6 @@ Listings:\n\
   -l, --list[=FORMAT]        list one or all known charsets and aliases\n\
   -k, --known=PAIRS          restrict charsets according to known PAIRS list\n\
   -h, --header[=[LN/]NAME]   write table NAME on stdout using LN, then exit\n\
-  -F, --freeze-tables        write out a C module holding all tables\n\
   -T, --find-subsets         report all charsets being subset of others\n\
   -C, --copyright            display Copyright and copying conditions\n\
       --help                 display this help and exit\n\
@@ -267,26 +231,14 @@ Operation modes:\n\
   -q, --quiet, --silent   inhibit messages about irreversible recodings\n\
   -f, --force             force recodings even when not reversible\n\
   -t, --touch             touch the recoded files after replacement\n\
-  -i, --sequence=files    use intermediate files for sequencing passes\n\
-      --sequence=memory   use memory buffers for sequencing passes\n\
+  -i, -p, --sequence=STRATEGY  ignored for backwards compatibility\n\
 "),
 	     stdout);
-#if HAVE_PIPE
-      fputs (_("\
-  -p, --sequence=pipe     use pipe machinery for sequencing passes\n\
-"),
-	     stdout);
-#else
-      fputs (_("\
-  -p, --sequence=pipe     same as -i (on this system)\n\
-"),
-	     stdout);
-#endif
       fputs (_("\
 \n\
 Fine tuning:\n\
   -s, --strict           use strict mappings, even loose characters\n\
-  -d, --diacritics       convert only diacritics or alike for HTML/LaTeX\n\
+  -d, --diacritics       convert only diacritics or alike for HTML/LaTeX/BibTeX\n\
   -S, --source[=LN]      limit recoding to strings and comments as for LN\n\
   -c, --colons           use colons instead of double quotes for diaeresis\n\
   -g, --graphics         approximate IBMPC rulers by ASCII graphics\n\
@@ -322,13 +274,12 @@ with an empty surface name means no surfaces at all.  See the manual.\n"),
 	     stdout);
       fputs (_("\
 \n\
-If none of -i and -p are given, presume -p if no FILE, else -i.\n\
 Each FILE is recoded over itself, destroying the original.  If no\n\
 FILE is specified, then act as a filter and recode stdin to stdout.\n"),
 	     stdout);
       fputs (_("\
 \n\
-Report bugs to <recode-bugs@iro.umontreal.ca>.\n"),
+Report bugs at https://github.com/rrthomas/Recode\n"),
 	       stdout);
     }
   exit (status);
@@ -347,7 +298,6 @@ static const struct option long_options[] =
   {"diacritics", no_argument, NULL, 'd'},
   {"find-subsets", no_argument, NULL, 'T'},
   {"force", no_argument, NULL, 'f'},
-  {"freeze-tables", no_argument, NULL, 'F'},
   {"header", optional_argument, NULL, 'h'},
   {"help", no_argument, &show_help, 1},
   {"ignore", required_argument, NULL, 'x'},
@@ -379,7 +329,7 @@ main (int argc, char *const *argv)
   int option_char;		/* option character */
   bool success = true;		/* reversibility of all recodings */
 
-  static bool (*processor) PARAMS ((RECODE_TASK));
+  static bool (*processor) (RECODE_TASK);
   struct recode_outer outer_option;
   struct recode_request request_option;
   struct recode_task task_option;
@@ -394,7 +344,6 @@ main (int argc, char *const *argv)
     program_name += 3;
 
   setlocale (LC_ALL, "");
-  bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
   /* Decode command options.  */
@@ -408,7 +357,6 @@ main (int argc, char *const *argv)
   memset (&task_option, 0, sizeof (struct recode_task));
 
   request_option.diaeresis_char = '"';
-  task_option.strategy = RECODE_STRATEGY_UNDECIDED;
   task_option.fail_level = RECODE_AMBIGUOUS_OUTPUT;
   task_option.abort_level = RECODE_AMBIGUOUS_OUTPUT;
 
@@ -429,22 +377,21 @@ main (int argc, char *const *argv)
 	  case -2:
 	    error (0, 0, _("Sequence `%s' is ambiguous"), optarg);
 	    usage (EXIT_FAILURE, 0);
+            break;
 
 	  case -1:
 	    error (0, 0, _("Sequence `%s' is unknown"), optarg);
 	    usage (EXIT_FAILURE, 0);
+            break;
 
+            /* Ignore for backwards compatibility with version 3.6.  */
 	  case 0:
-	    task_option.strategy = RECODE_SEQUENCE_IN_MEMORY;
-	    break;
-
 	  case 1:
-	    task_option.strategy = RECODE_SEQUENCE_WITH_FILES;
+	  case 2:
 	    break;
 
-	  case 2:
-	    task_option.strategy = RECODE_SEQUENCE_WITH_PIPE;
-	    break;
+          default:
+            break;
 	  }
 	break;
 
@@ -452,10 +399,6 @@ main (int argc, char *const *argv)
 	print_copyright ();
 	exit (EXIT_SUCCESS);
 
-      case 'F':
-	freeze_tables = true;
-	break;
-
       case 'S':
 	if (optarg)
 	  switch (argmatch (optarg, language_strings, NULL, 0))
@@ -463,10 +406,12 @@ main (int argc, char *const *argv)
 	    case -2:
 	      error (0, 0, _("Language `%s' is ambiguous"), optarg);
 	      usage (EXIT_FAILURE, 0);
+              break;
 
 	    default:		/* -1 */
 	      error (0, 0, _("Language `%s' is unknown"), optarg);
 	      usage (EXIT_FAILURE, 0);
+              break;
 
 	    case 0:
 	      processor = transform_c_source;
@@ -521,10 +466,12 @@ main (int argc, char *const *argv)
 	      case -2:
 		error (0, 0, _("Language `%s' is ambiguous"), buffer);
 		usage (EXIT_FAILURE, 0);
+                break;
 
 	      default:		/* -1 */
 		error (0, 0, _("Language `%s' is unknown"), buffer);
 		usage (EXIT_FAILURE, 0);
+                break;
 
 	      case 0:
 		header_language = RECODE_LANGUAGE_C;
@@ -544,7 +491,7 @@ main (int argc, char *const *argv)
 	break;
 
       case 'i':
-	task_option.strategy = RECODE_SEQUENCE_WITH_FILES;
+        /* Ignore for backwards compatibility with version 3.6.  */
 	break;
 
       case 'k':
@@ -559,10 +506,12 @@ main (int argc, char *const *argv)
 	    case -2:
 	      error (0, 0, _("Format `%s' is ambiguous"), optarg);
 	      usage (EXIT_FAILURE, 0);
+              break;
 
 	    case -1:
 	      error (0, 0, _("Format `%s' is unknown"), optarg);
 	      usage (EXIT_FAILURE, 0);
+              break;
 
 	    case 0:
 	      list_format = RECODE_DECIMAL_FORMAT;
@@ -579,11 +528,14 @@ main (int argc, char *const *argv)
 	    case 3:
 	      list_format = RECODE_FULL_FORMAT;
 	      break;
+
+            default:
+              break;
 	    }
 	break;
 
       case 'p':
-	task_option.strategy = RECODE_SEQUENCE_WITH_PIPE;
+        /* Ignore for backwards compatibility with version 3.6.  */
 	break;
 
       case 'q':
@@ -617,13 +569,13 @@ main (int argc, char *const *argv)
 
   if (show_version)
     {
-      printf ("Free %s %s\n", PACKAGE, VERSION);
+      printf ("%s %s\n", PACKAGE, VERSION);
       fputs (_("\
-Written by Franc,ois Pinard <pinard@iro.umontreal.ca>.\n"),
+Written by François Pinard <pinard@iro.umontreal.ca>.\n"),
 	     stdout);
       fputs (_("\
 \n\
-Copyright (C) 1990, 92, 93, 94, 96, 97, 99 Free Software Foundation, Inc.\n"),
+Copyright (C) 1990, 92-94, 96, 97, 99, 08 Free Software Foundation, Inc.\n"),
 	     stdout);
       fputs (_("\
 This is free software; see the source for copying conditions.  There is NO\n\
@@ -637,15 +589,16 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 
   /* Register all modules and build internal tables.  */
 
-  outer = recode_new_outer (true);
-  if (!outer)
-    abort ();
-
-  if (freeze_tables)
-    {
-      recode_freeze_tables (outer);
-      exit (EXIT_SUCCESS);
-    }
+  {
+    unsigned flags = RECODE_AUTO_ABORT_FLAG;
+    if (ignored_name && *ignored_name == ':')
+      flags |= RECODE_NO_ICONV_FLAG;
+    if (request_option.make_header_flag)
+      flags |= RECODE_NO_ICONV_FLAG;
+    outer = recode_new_outer (flags);
+    if (!outer)
+      abort ();
+  }
 
   /* Set strict mapping.  */
 
@@ -801,8 +754,7 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 
     if (!recode_scan_request (request, user_request))
       error (EXIT_FAILURE, 0, _("Request `%s' is erroneous"), user_request);
-    if (rewritten_request)
-      free (rewritten_request);
+    free (rewritten_request);
   }
 
   /* If we merely want source code, do it and get out.  */
@@ -813,13 +765,14 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
       exit (EXIT_SUCCESS);
     }
 
+#if HAVE_PIPE
   setup_signals ();
+#endif
 
   {
     RECODE_TASK task;
 
     task = recode_new_task (request);
-    task->strategy = task_option.strategy;
     task->fail_level = task_option.fail_level;
     task->abort_level = task_option.fail_level;
 
@@ -828,15 +781,6 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 
     if (optind < argc)
       {
-	/* When reading and writing files, unless the user selected
-	   otherwise, avoid forking and use intermediate files.  */
-
-	/* FIXME: On a file per file basis, force recoding to be done in
-	   memory whenever files are small, which is the usual case.  */
-
-	if (task->strategy == RECODE_STRATEGY_UNDECIDED)
-	  task->strategy = RECODE_SEQUENCE_WITH_FILES;
-
 	/* In case files are recoded over themselves and there is no
 	   recoding step at all, do not even try to touch the files.  */
 
@@ -847,56 +791,43 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 	  for (; optind < argc; optind++)
 	    {
 	      const char *input_name;
-	      char output_name[200]; /* FIXME: dangerous limit */
 	      FILE *file;
 	      struct stat file_stat;
 	      struct utimbuf file_utime;
 
-	      input_name = argv[optind];
+	      input_name = realpath (argv[optind], NULL);
+	      if (input_name == NULL)
+		error (EXIT_FAILURE, errno, "realpath (%s)", argv[optind]);
 
 	      /* Check if the file can be read and rewritten.  */
 
 	      if (file = fopen (input_name, "r+"), file == NULL)
 		error (EXIT_FAILURE, errno, "fopen (%s)", input_name);
 
-	      /* Save the input file attrobites.  */
+	      /* Save the input file attributes.  */
 
 	      fstat (fileno (file), &file_stat);
 	      fclose (file);
 
-	      {
-		char *cursor;
+	      /* Choose an output file in the same directory.  */
 
-		/* Choose an output file in the same directory.  */
-
-		/* FIXME: Scott Schwartz <schwartz@bio.cse.psu.edu> writes:
-		   "There's no reason to think that that name is unique."  */
-
-		strcpy (output_name, input_name);
-#if DOSWIN_OR_OS2
-		for (cursor = output_name + strlen (output_name);
-		     cursor > output_name && cursor[-1] != '/'
-		       && cursor[-1] != '\\' && cursor[-1] != ':';
-		     cursor--)
-		  ;
-# if __DJGPP__
-		sprintf (cursor, "rec%d.tmp", getpid ());
-# else
-		strcpy (cursor, "recodeXX.TMP");
-# endif
-#else
-		for (cursor = output_name + strlen (output_name);
-		     cursor > output_name && cursor[-1] != '/';
-		     cursor--)
-		  ;
-		sprintf (cursor, "rec%d.tmp", getpid ());
-#endif
-	      }
+	      char *input_name_copy = xstrdup (input_name);
+	      char *output_dir = dirname (input_name_copy);
+	      char *output_name;
+	      if (asprintf (&output_name, "%s/recode-XXXXXX.tmp", output_dir) == -1)
+		error (EXIT_FAILURE, errno, "asprintf");
+	      int fd = mkstemps (output_name, 4);
+	      xset_binary_mode (fd, O_BINARY);
+	      if (fd == -1)
+		error (EXIT_FAILURE, errno, "mkstemps (%s)", output_name);
 
 	      /* Recode the file.  */
 
 	      task->input.name = input_name;
-	      task->output.name = output_name;
+	      task->output.name = NULL;
+	      task->output.file = fdopen (fd, "w+");
+	      if (task->output.file == NULL)
+	      	error (EXIT_FAILURE, errno, "fdopen ()");
 
 	      if (verbose_flag)
 		{
@@ -919,21 +850,13 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 		  if (unlink (input_name) < 0)
 		    error (EXIT_FAILURE, errno, "unlink (%s)", input_name);
 
-		  /* Preserve the file permissions.  */
+		  /* Preserve the file permissions if possible.  */
+
+		  chmod (output_name, file_stat.st_mode & 07777);
 
-		  if (chmod (output_name, file_stat.st_mode & 07777) < 0)
-		    error (EXIT_FAILURE, errno, "chmod (%s)", output_name);
-#if HAVE_RENAME
 		  if (rename (output_name, input_name) < 0)
 		    error (EXIT_FAILURE, errno, "rename (%s, %s)",
 			   output_name, input_name);
-#else
-		  if (link (output_name, input_name) < 0)
-		    error (EXIT_FAILURE, errno, "link (%s, %s)",
-			   output_name, input_name);
-		  if (unlink (output_name) < 0)
-		    error (EXIT_FAILURE, errno, "unlink (%s)", output_name);
-#endif
 
 		  /* Adjust the time stamp for the new file.  */
 
@@ -965,16 +888,12 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
 
 		  unlink (output_name);
 		}
+	      free (output_name);
+	      free (input_name_copy);
 	    }
       }
     else
       {
-	/* When reading stdin and writing stdout, unless the user selected
-	   otherwise, fork processes interconnected with pipes.  */
-
-	if (task->strategy == RECODE_STRATEGY_UNDECIDED)
-	  task->strategy = RECODE_SEQUENCE_WITH_PIPE;
-
 	task->input.name = "";
 	task->output.name = "";
 	if (!(*processor) (task))
