@@ -1,10 +1,10 @@
 /* Conversion of files between different charsets and surfaces.
-   Copyright  1996, 97, 98, 99, 00 Free Software Foundation, Inc.
-   Contributed by Franois Pinard <pinard@iro.umontreal.ca>, 1996.
+   Copyright © 1996, 97, 98, 99, 00 Free Software Foundation, Inc.
+   Contributed by François Pinard <pinard@iro.umontreal.ca>, 1996.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
-   as published by the Free Software Foundation; either version 2 of the
+   as published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be
@@ -13,11 +13,13 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the `recode' Library; see the file `COPYING.LIB'.
+   License along with the Recode Library; see the file `COPYING.LIB'.
    If not, write to the Free Software Foundation, Inc., 59 Temple Place -
    Suite 330, Boston, MA 02111-1307, USA.  */
 
 #include "common.h"
+#include "decsteps.h"
+
 static const bool bitnet_flag = false;
 
 /* Maximum number of characters per MIME line.  */
@@ -55,9 +57,9 @@ static const char safe_char_bitnet[1 << 7] =
     unsigned nibble;							\
 								\
     put_byte ('=', subtask);					\
-    nibble = MASK (4) & (Character) >> 4;			\
+    nibble = BIT_MASK (4) & (Character) >> 4;			\
     put_byte ((nibble < 10 ? '0' : 'A' - 10) + nibble, subtask);	\
-    nibble = MASK (4) & (Character);				\
+    nibble = BIT_MASK (4) & (Character);				\
     put_byte ((nibble < 10 ? '0' : 'A' - 10) + nibble, subtask);	\
   } while (false)
 
@@ -358,7 +360,7 @@ module_quoted_printable (RECODE_OUTER outer)
     && declare_alias (outer, "QP", "Quoted-Printable");
 }
 
-void
-delmodule_quoted_printable (RECODE_OUTER outer)
+_GL_ATTRIBUTE_CONST void
+delmodule_quoted_printable (RECODE_OUTER outer _GL_UNUSED_PARAMETER)
 {
 }
