@@ -1,10 +1,10 @@
 /* Conversion of files between different charsets and surfaces.
-   Copyright  1990,92,93,94,96,97,98,99,00 Free Software Foundation, Inc.
-   Contributed by Franois Pinard <pinard@iro.umontreal.ca>, 1990.
+   Copyright © 1990,92,93,94,96,97,98,99,00 Free Software Foundation, Inc.
+   Contributed by François Pinard <pinard@iro.umontreal.ca>, 1990.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
-   as published by the Free Software Foundation; either version 2 of the
+   as published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be
@@ -13,7 +13,7 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the `recode' Library; see the file `COPYING.LIB'.
+   License along with the Recode Library; see the file `COPYING.LIB'.
    If not, write to the Free Software Foundation, Inc., 59 Temple Place -
    Suite 330, Boston, MA 02111-1307, USA.  */
 
@@ -21,15 +21,9 @@
 
 /* Global declarations and definitions.  */
 
+#include <stdio.h>
 #include <ctype.h>
-
-#if HAVE_UNISTD_H
-# include <unistd.h>
-#endif
-
-#if WITH_DMALLOC
-# include <dmalloc.h>
-#endif
+#include <unistd.h>
 
 #include "hash.h"
 
@@ -37,70 +31,29 @@ extern const char *program_name;
 
 /* Error handling.  */
 
-#if HAVE_VPRINTF || HAVE_DOPRNT
-# if __STDC__
-#  include <stdarg.h>
-#  define VA_START(args, lastarg) va_start(args, lastarg)
-# else
-#  include <varargs.h>
-#  define VA_START(args, lastarg) va_start(args)
-# endif
-#else
-# define va_alist a1, a2, a3, a4, a5, a6, a7, a8
-# define va_dcl char *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
-#endif
+#include <stdarg.h>
 
-void
-#if defined(VA_START) && __STDC__
-recode_error (RECODE_OUTER outer, const char *format, ...)
-#else
-recode_error (outer, format, va_alist)
-     RECODE_OUTER outer;
-     const char *format;
-     va_dcl
-#endif
+_GL_ATTRIBUTE_FORMAT_PRINTF (2, 3) void
+recode_error (RECODE_OUTER outer _GL_UNUSED_PARAMETER, const char *format, ...)
 {
-#ifdef VA_START
   va_list args;
 
-  VA_START (args, format);
-# if HAVE_VPRINTF
+  va_start (args, format);
   vfprintf (stderr, format, args);
-# else
-  _doprnt (format, args, stderr);
-# endif
   va_end (args);
-#else
-  fprintf (stderr, format, a1, a2, a3, a4, a5, a6, a7, a8);
-#endif
   putc ('\n', stderr);
   fflush (stderr);
 }
 
-void
-#if defined(VA_START) && __STDC__
-recode_perror (RECODE_OUTER outer, const char *format, ...)
-#else
-recode_perror (outer, format, va_alist)
-     RECODE_OUTER outer;
-     const char *format;
-     va_dcl
-#endif
+_GL_ATTRIBUTE_FORMAT_PRINTF (2, 3) void
+recode_perror (RECODE_OUTER outer _GL_UNUSED_PARAMETER, const char *format, ...)
 {
   int saved_errno = errno;
-#ifdef VA_START
   va_list args;
 
-  VA_START (args, format);
-# if HAVE_VPRINTF
+  va_start (args, format);
   vfprintf (stderr, format, args);
-# else
-  _doprnt (format, args, stderr);
-# endif
   va_end (args);
-#else
-  fprintf (stderr, format, a1, a2, a3, a4, a5, a6, a7, a8);
-#endif
   fprintf (stderr, ": %s\n", strerror (saved_errno));
   fflush (stderr);
 }
@@ -110,7 +63,7 @@ recode_malloc (RECODE_OUTER outer, size_t size)
 {
   void *result;
 
-  result = malloc (size);
+  result = calloc (1, size);
   if (!result)
     recode_error (outer, _("Virtual memory exhausted"));
 
@@ -152,7 +105,7 @@ invert_table (RECODE_OUTER outer, const unsigned char *table)
     {
       if (flag[table[counter]])
 	{
-	  recode_error (outer, _("Codes %3d and %3d both recode to %3d"),
+	  recode_error (outer, _("Codes %3d and %3u both recode to %3d"),
 			result[table[counter]], counter, table[counter]);
 	  table_error = true;
 	}
@@ -166,7 +119,7 @@ invert_table (RECODE_OUTER outer, const unsigned char *table)
     {
       for (counter = 0; counter < 256; counter++)
 	if (!flag[counter])
-	  recode_error (outer, _("No character recodes to %3d"), counter);
+	  recode_error (outer, _("No character recodes to %3u"), counter);
       recode_error (outer, _("Cannot invert given one-to-one table"));
     }
   return result;
@@ -206,6 +159,8 @@ complete_pairs (RECODE_OUTER outer, RECODE_STEP step,
 
   memset (left_flag, 0, 256);
   memset (right_flag, 0, 256);
+  memset (left_table, 0, 256);
+  memset (right_table, 0, 256);
   table_error = false;
 
   /* Establish known data.  */
@@ -226,7 +181,7 @@ complete_pairs (RECODE_OUTER outer, RECODE_STEP step,
 	      table_error = true;
 	    }
 	  recode_error (outer,
-			_("Pair no. %d: <%3d, %3d> conflicts with <%3d, %3d>"),
+			_("Pair no. %u: <%3d, %3d> conflicts with <%3d, %3d>"),
 			counter, left, right, left, left_table[left]);
 	}
       else if (right_flag[right])
@@ -238,7 +193,7 @@ complete_pairs (RECODE_OUTER outer, RECODE_STEP step,
 	      table_error = true;
 	    }
 	  recode_error (outer,
-			_("Pair no. %d: <%3d, %3d> conflicts with <%3d, %3d>"),
+			_("Pair no. %u: <%3d, %3d> conflicts with <%3d, %3d>"),
 			counter, left, right, right_table[right], right);
 	}
       else
@@ -293,6 +248,7 @@ complete_pairs (RECODE_OUTER outer, RECODE_STEP step,
       memcpy (table, reverse ? right_table : left_table, 256);
       step->step_type = RECODE_BYTE_TO_BYTE;
       step->step_table = table;
+      step->step_table_term_routine = free;
 
       /* Upgrade step quality to reversible.  */
 
@@ -344,6 +300,7 @@ complete_pairs (RECODE_OUTER outer, RECODE_STEP step,
       step->transform_routine = transform_byte_to_variable;
       step->step_type = RECODE_BYTE_TO_STRING;
       step->step_table = table2;
+      step->step_table_term_routine = free;
     }
 
   return true;
@@ -388,10 +345,22 @@ transform_byte_to_ucs2 (RECODE_SUBTASK subtask)
 | Recode a file from double byte UCS-2 characters to one byte characters.  |
 `-------------------------------------------------------------------------*/
 
-static unsigned
-ucs2_to_byte_hash (const void *void_data, unsigned table_size)
+struct ucs2_to_byte
+  {
+    recode_ucs2 code;		/* UCS-2 value */
+    unsigned char byte;		/* corresponding byte */
+  };
+
+struct ucs2_to_byte_local
+  {
+    Hash_table *table;
+    struct ucs2_to_byte *data;
+  };
+
+static size_t
+ucs2_to_byte_hash (const void *void_data, size_t table_size)
 {
-  const struct ucs2_to_byte *data = void_data;
+  const struct ucs2_to_byte *data = (const struct ucs2_to_byte *) void_data;
 
   return data->code % table_size;
 }
@@ -399,12 +368,21 @@ ucs2_to_byte_hash (const void *void_data, unsigned table_size)
 static bool
 ucs2_to_byte_compare (const void *void_first, const void *void_second)
 {
-  const struct ucs2_to_byte *first = void_first;
-  const struct ucs2_to_byte *second = void_second;
+  const struct ucs2_to_byte *first = (const struct ucs2_to_byte *) void_first;
+  const struct ucs2_to_byte *second = (const struct ucs2_to_byte *) void_second;
 
   return first->code == second->code;
 }
 
+static bool
+term_ucs2_to_byte (RECODE_STEP step)
+{
+  hash_free (((struct ucs2_to_byte_local *) step->local)->table);
+  free (((struct ucs2_to_byte_local *) step->local)->data);
+  free (step->local);
+  return true;
+}
+
 bool
 init_ucs2_to_byte (RECODE_STEP step,
 		   RECODE_CONST_REQUEST request,
@@ -442,14 +420,23 @@ init_ucs2_to_byte (RECODE_STEP step,
 	}
     }
 
-  step->local = table;
+  if (!ALLOC (step->local, 1, struct ucs2_to_byte_local))
+    {
+      hash_free (table);
+      free (data);
+      return false;
+    }
+  ((struct ucs2_to_byte_local *) step->local)->table = table;
+  ((struct ucs2_to_byte_local *) step->local)->data = data;
+
+  step->term_routine = term_ucs2_to_byte;
   return true;
 }
 
 bool
 transform_ucs2_to_byte (RECODE_SUBTASK subtask)
 {
-  Hash_table *table = subtask->step->local;
+  Hash_table *table = ((struct ucs2_to_byte_local *) subtask->step->local)->table;
   struct ucs2_to_byte lookup;
   struct ucs2_to_byte *entry;
   unsigned input_value;		/* current UCS-2 character */
@@ -457,7 +444,7 @@ transform_ucs2_to_byte (RECODE_SUBTASK subtask)
   while (get_ucs2 (&input_value, subtask))
     {
       lookup.code = input_value;
-      entry = hash_lookup (table, &lookup);
+      entry = (struct ucs2_to_byte *) hash_lookup (table, &lookup);
       if (entry)
 	put_byte (entry->byte, subtask);
       else
@@ -533,7 +520,7 @@ recode_format_table (RECODE_REQUEST request,
 
   /* Print the header of the header file.  */
 
-  printf (_("%sConversion table generated mechanically by Free `%s' %s"),
+  printf (_("%sConversion table generated mechanically by %s %s"),
 	  start_comment, PACKAGE, VERSION);
   printf (_("%sfor sequence %s.%s"),
 	  wrap_comment, edit_sequence (request, 1), end_comment);
@@ -575,7 +562,7 @@ recode_format_table (RECODE_REQUEST request,
 
   if (step->step_type == RECODE_BYTE_TO_BYTE)
     {
-      const unsigned char *table = step->step_table;
+      const unsigned char *table = (const unsigned char *) step->step_table;
 
       /* Produce a one to one recoding table.  */
 
@@ -593,12 +580,15 @@ recode_format_table (RECODE_REQUEST request,
 	  printf ("@%s =\n", name);
 	  printf ("  (\n");
 	  break;
+
+        default:
+          break;
 	}
       for (counter = 0; counter < 256; counter++)
 	{
 	  printf ("%s%3d,", counter % 8 == 0 ? "    " : " ", table[counter]);
 	  if (counter % 8 == 7)
-	    printf ("\t%s%3d - %3d%s",
+	    printf ("\t%s%3u - %3u%s",
 		    start_comment, counter - 7, counter, end_comment);
 	}
       switch (header_language)
@@ -613,11 +603,14 @@ recode_format_table (RECODE_REQUEST request,
 	case RECODE_LANGUAGE_PERL:
 	  printf ("  );\n");
 	  break;
+
+        default:
+          break;
 	}
     }
   else if (step->step_type == RECODE_BYTE_TO_STRING)
     {
-      const char *const *table = step->step_table;
+      const char *const *table = (const char *const *) step->step_table;
 
       /* Produce a one to many recoding table.  */
 
@@ -635,6 +628,9 @@ recode_format_table (RECODE_REQUEST request,
 	  printf ("@%s =\n", name);
 	  printf ("  (\n");
 	  break;
+
+        default:
+          break;
 	}
       for (counter = 0; counter < 256; counter++)
 	{
@@ -684,6 +680,7 @@ recode_format_table (RECODE_REQUEST request,
 			column += 2;
 			break;
 		      }
+                    FALLTHROUGH;
 
 		  default:
 		    if (isprint (*cursor2))
@@ -715,6 +712,9 @@ recode_format_table (RECODE_REQUEST request,
 		printf ("''");
 		column += 2;
 		break;
+
+              default:
+                break;
 	      }
 	  printf (",");
 	  column++;
@@ -723,7 +723,7 @@ recode_format_table (RECODE_REQUEST request,
 	      printf ("\t");
 	      column += 8 - column % 8;
 	    }
-	  printf ("%s%3d%s", start_comment, counter, end_comment);
+	  printf ("%s%3u%s", start_comment, counter, end_comment);
 	}
       switch (header_language)
 	{
@@ -737,6 +737,9 @@ recode_format_table (RECODE_REQUEST request,
 	case RECODE_LANGUAGE_PERL:
 	  printf ("  );\n");
 	  break;
+
+        default:
+          break;
 	}
     }
   else
