@@ -1,10 +1,10 @@
 /* Conversion of files between different charsets and surfaces.
-   Copyright  1990,92,93,94,96,97,98,99,00 Free Software Foundation, Inc.
-   Contributed by Franois Pinard <pinard@iro.umontreal.ca>, 1990.
+   Copyright © 1990,92,93,94,96,97,98,99,00,01 Free Software Foundation, Inc.
+   Contributed by François Pinard <pinard@iro.umontreal.ca>, 1990.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
-   as published by the Free Software Foundation; either version 2 of the
+   as published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be
@@ -13,7 +13,7 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with the `recode' Library; see the file `COPYING.LIB'.
+   License along with the Recode Library; see the file `COPYING.LIB'.
    If not, write to the Free Software Foundation, Inc., 59 Temple Place -
    Suite 330, Boston, MA 02111-1307, USA.  */
 
@@ -73,14 +73,11 @@ add_work_character (RECODE_REQUEST request, int character)
       new_work_string = (char *)
 	realloc (request->work_string, request->work_string_allocated);
       if (new_work_string)
-	{
-	  request->work_string = new_work_string;
-	  request->work_string[request->work_string_length++] = character;
-	}
-      /* else, the diagnostic gets truncated, no need to fuss about it.  */
+        request->work_string = new_work_string;
+      else
+        return; /* the diagnostic gets truncated, no need to fuss about it.  */
     }
-  else
-    request->work_string[request->work_string_length++] = character;
+  request->work_string[request->work_string_length++] = character;
 }
 
 /*----------------------------------.
@@ -124,8 +121,7 @@ edit_sequence (RECODE_REQUEST request, bool edit_quality)
 	  /* Find unsurfacers.  */
 
 	  while (step < request->sequence_array + request->sequence_length
-		 && (step->after == outer->data_symbol
-		     || step->after == outer->tree_symbol))
+		 && step->after == outer->data_symbol)
 	    step++;
 	  unsurfacer_end = step;
 
@@ -141,12 +137,6 @@ edit_sequence (RECODE_REQUEST request, bool edit_quality)
 		{
 		  last_charset_printed = step->before;
 		  add_work_string (request, last_charset_printed->name);
-#if 0
-		  if (unsurfacer_start == unsurfacer_end
-		      && last_charset_printed
-		      && last_charset_printed->implied_surfaces)
-		    add_work_character (request, '/');
-#endif
 		}
 	    }
 
@@ -163,33 +153,22 @@ edit_sequence (RECODE_REQUEST request, bool edit_quality)
 
 	  add_work_string (request, "..");
 	  if (step < request->sequence_array + request->sequence_length
-	      && step->before != outer->data_symbol
-	      && step->before != outer->tree_symbol)
+	      && step->before != outer->data_symbol)
 	    {
 	      last_charset_printed = step->after;
 	      add_work_string (request, last_charset_printed->name);
 	      step++;
-#if 0
-	      if ((step == request->sequence_array + request->sequence_length
-		   || (step->before != outer->data_symbol
-		       && step->before != outer->tree_symbol))
-		  && last_charset_printed
-		  && last_charset_printed->implied_surfaces)
-		add_work_character (request, '/');
-#endif
 	    }
 	  else
 	    {
 	      last_charset_printed = outer->data_symbol;
-	      /* FIXME: why not outer->tree_symbol?  */
 	      add_work_string (request, last_charset_printed->name);
 	    }
 
 	  /* Print resurfacers.  */
 
 	  while (step < request->sequence_array + request->sequence_length
-		 && (step->before == outer->data_symbol
-		     || step->before == outer->tree_symbol))
+		 && step->before == outer->data_symbol)
 	    {
 	      add_work_character (request, '/');
 	      last_charset_printed = NULL;
@@ -233,8 +212,10 @@ add_to_sequence (RECODE_REQUEST request, RECODE_SINGLE single,
   RECODE_OUTER outer = request->outer;
   RECODE_STEP step;
 
-  if (request->sequence_length == request->sequence_allocated)
+  if ((unsigned)request->sequence_length == request->sequence_allocated)
     {
+      unsigned old_allocated = request->sequence_allocated;
+
       if (request->sequence_allocated == 0)
 	request->sequence_allocated = 16;
       else
@@ -246,16 +227,21 @@ add_to_sequence (RECODE_REQUEST request, RECODE_SINGLE single,
 	  recode_error (outer, _("Virtual memory exhausted!"));
 	  return false;
 	}
+
+      memset (request->sequence_array + old_allocated, 0,
+              (request->sequence_allocated - old_allocated) * sizeof (struct recode_step));
     }
 
   step = request->sequence_array + request->sequence_length++;
   step->before = single->before;
   step->after = single->after;
   step->step_table = single->initial_step_table;
+  step->step_table_term_routine = NULL;
   step->step_type
     = step->step_table ? RECODE_COMBINE_EXPLODE : RECODE_NO_STEP_TABLE;
   step->transform_routine = single->transform_routine;
   step->fallback_routine = single->fallback_routine;
+  step->term_routine = NULL;
 
   if (single->init_routine)
     {
@@ -472,7 +458,7 @@ complete_double_ucs2_step (RECODE_OUTER outer, RECODE_STEP step)
     {
       /* Construct the array of binding items for the charset.  */
 
-      data = side->charset->data;
+      data = (const struct strip_data *) side->charset->data;
       pool = data->pool;
       item_cursor = side->item;
       byte = 0;
@@ -487,7 +473,7 @@ complete_double_ucs2_step (RECODE_OUTER outer, RECODE_STEP step)
 	    {
 	      unsigned code = pool[offset + position_counter];
 
-	      if (code != MASK (16))
+	      if (code != BIT_MASK (16))
 		{
 		  /* Establish a new binding item.  */
 		  item_cursor->byte = byte;
@@ -543,6 +529,24 @@ complete_double_ucs2_step (RECODE_OUTER outer, RECODE_STEP step)
 		    pair_array, pair_cursor - pair_array, false, reversed);
 }
 
+static bool
+delete_compressed_one_to_many (RECODE_STEP step)
+{
+  free (step->local);
+  return true;
+}
+
+static void
+delete_step (RECODE_STEP step)
+{
+  /* Run step destructor, and delete step_table if allocated.  */
+  if (step->term_routine)
+    (*step->term_routine) (step);
+
+  if (step->step_table_term_routine != NULL)
+    (*step->step_table_term_routine) (step->step_table);
+}
+
 /*---------------------------------------------------------------------.
 | Optimize a SEQUENCE of single steps by creating new single steps, if |
 | this can be done by merging adjacent steps which are simple enough.  |
@@ -554,9 +558,9 @@ simplify_sequence (RECODE_REQUEST request)
   RECODE_OUTER outer = request->outer;
 
   unsigned saved_steps;		/* number of saved steps */
-  RECODE_STEP in;	/* next studied sequence step */
-  RECODE_STEP out;	/* next rewritten sequence step */
-  RECODE_STEP limit;	/* last value for IN */
+  RECODE_STEP in;               /* next studied sequence step */
+  RECODE_STEP out;              /* next rewritten sequence step */
+  RECODE_STEP limit;            /* last value for IN */
   unsigned char *accum;		/* byte_to_byte accumulated recoding */
   const char **string;		/* byte_to_variable recoding */
   unsigned char temp[256];	/* temporary value for accum array */
@@ -582,6 +586,10 @@ simplify_sequence (RECODE_REQUEST request)
 	&& in[1].before == outer->ucs2_charset
 	&& in[1].after->data_type == RECODE_STRIP_DATA)
       {
+        /* Free old steps before overwriting anything.  */
+        delete_step (in);
+        delete_step (in + 1);
+
 	/* This is a double UCS-2 step.  */
 	out->before = in[0].before;
 	out->after = in[1].after;
@@ -599,15 +607,19 @@ simplify_sequence (RECODE_REQUEST request)
 	out++;
       }
     else if (in < limit - 1
-	     && in[0].after == outer->libiconv_pivot
-	     && in[1].before == outer->libiconv_pivot)
+	     && in[0].after == outer->iconv_pivot
+	     && in[1].before == outer->iconv_pivot)
       {
-	/* This is a double `libiconv' step.  */
+        /* Free old steps before overwriting anything.  */
+        delete_step (in);
+        delete_step (in + 1);
+
+	/* This is a double `iconv' step.  */
 	out->before = in[0].before;
 	out->after = in[1].after;
 	out->quality = in[0].quality;
 	merge_qualities (&out->quality, in[1].quality);
-	out->transform_routine = transform_with_libiconv;
+	out->transform_routine = transform_with_iconv;
 
 	in += 2;
 	saved_steps++;
@@ -642,14 +654,14 @@ simplify_sequence (RECODE_REQUEST request)
 	out->before = in->before;
 	out->after = in->after;
 	out->quality = in->quality;
-	in++;
+	delete_step (in++);
 
 	/* Merge in all consecutive one-to-one recodings.  */
 
 	while (in < limit
 	       && (table_type (request, in) == RECODE_BYTE_TO_BYTE))
 	  {
-	    const unsigned char *table = in->step_table;
+	    const unsigned char *table = (const unsigned char *) in->step_table;
 
 	    for (counter = 0; counter < 256; counter++)
 	      temp[counter] = table[accum[counter]];
@@ -657,7 +669,7 @@ simplify_sequence (RECODE_REQUEST request)
 
 	    out->after = in->after;
 	    merge_qualities (&out->quality, in->quality);
-	    in++;
+	    delete_step (in++);
 	    saved_steps++;
 	  }
 
@@ -670,13 +682,19 @@ simplify_sequence (RECODE_REQUEST request)
 
 	    && (ALLOC (string, 256, const char *)))
 	  {
-	    const char *const *table = in->step_table;
+	    const char *const *table = (const char *const *) in->step_table;
 
 	    for (counter = 0; counter < 256; counter++)
 	      string[counter] = table[accum[counter]];
 	    free (accum);
 	    out->step_type = RECODE_BYTE_TO_STRING;
 	    out->step_table = string;
+            if (in->step_table_term_routine)
+              {
+                out->local = (void *) table;   /* Save reference to old table for destructor.  */
+                out->term_routine = delete_compressed_one_to_many;
+              }
+            out->step_table_term_routine = free;
 	    out->transform_routine = transform_byte_to_variable;
 	    out->after = in->after;
 	    merge_qualities (&out->quality, in->quality);
@@ -712,6 +730,7 @@ simplify_sequence (RECODE_REQUEST request)
     {
       request->sequence_length = 0;
       saved_steps++;
+      delete_step (in);
     }
 
   /* Tell the user if something changed.  */
@@ -721,7 +740,7 @@ simplify_sequence (RECODE_REQUEST request)
   return true;
 }
 
-/* Decoging step sequence parameters.  */
+/* Decoding step sequence parameters.  */
 
 /*-------------------------------------.
 | Scan next name into SCANNED_STRING.  |
@@ -741,9 +760,10 @@ scan_identifier (RECODE_REQUEST request)
   return true;
 }
 
-/*---------------------------------------------.
-| Scan all options and return a list of them.  |
-`---------------------------------------------*/
+/*----------------------------------------------------------------.
+| Scan all options and return a list of them.                     |
+| FIXME: Need to be able to differentiate empty list from error.  |
+`----------------------------------------------------------------*/
 
 static RECODE_OPTION_LIST
 scan_options (RECODE_REQUEST request)
@@ -754,25 +774,28 @@ scan_options (RECODE_REQUEST request)
 
   while (*request->scan_cursor == '+')
     {
-      RECODE_OPTION_LIST new
-	= ALLOC (new, 1, struct recode_option_list);
+      RECODE_OPTION_LIST new_
+	= ALLOC (new_, 1, struct recode_option_list);
       char *copy;
 
-      if (!new)
+      if (!new_)
 	break;			/* FIXME: should interrupt decoding */
 
       request->scan_cursor++;
       scan_identifier (request);
       ALLOC (copy, strlen (request->scanned_string) + 1, char);
       if (!copy)
-	break;			/* FIXME: should interrupt decoding */
+	{
+	  free (new_);
+	  break;		/* FIXME: should interrupt decoding */
+	}
       strcpy (copy, request->scanned_string);
 
-      new->option = copy;
+      new_->option = copy;
       if (!list)
-	list = new;
-      new->next = last;
-      last = new;
+	list = new_;
+      new_->next = last;
+      last = new_;
     }
   return list;
 }
@@ -1064,29 +1087,30 @@ decode_request (RECODE_REQUEST request, const char *string)
 `--------------------------*/
 
 /* Guarantee four NULs at the end of the output memory buffer for TASK, yet
-   not counting them as data.  The number of NULs should depend on the goal
-   charset: often one, but two for UCS-2 and four for UCS-4.  FIXME!  */
+   not counting them as data.  (Four is the maximum, needed for UCS-4.)  */
 
-static void
+static bool
 guarantee_nul_terminator (RECODE_TASK task)
 {
   if (task->output.cursor + 4 >= task->output.limit)
     {
       RECODE_OUTER outer = task->request->outer;
-      size_t old_size = task->output.limit - task->output.buffer;
-      size_t new_size = task->output.cursor + 4 - task->output.buffer;
+      size_t size = task->output.cursor - task->output.buffer;
 
-      /* FIXME: Rethink about how the error should be reported.  */
-      if (REALLOC (task->output.buffer, new_size, char))
+      if (REALLOC (task->output.buffer, size + 4, char))
 	{
-	  task->output.cursor = task->output.buffer + old_size;
-	  task->output.limit = task->output.buffer + new_size;
+	  task->output.cursor = task->output.buffer + size;
+	  task->output.limit = task->output.buffer + size + 4;
 	}
+      else
+        return false;
     }
   task->output.cursor[0] = NUL;
   task->output.cursor[1] = NUL;
   task->output.cursor[2] = NUL;
   task->output.cursor[3] = NUL;
+
+  return true;
 }
 
 RECODE_REQUEST
@@ -1097,7 +1121,6 @@ recode_new_request (RECODE_OUTER outer)
   if (!ALLOC (request, 1, struct recode_request))
     return NULL;
 
-  memset (request, 0, sizeof (struct recode_request));
   request->outer = outer;
   request->diaeresis_char = '"';
 
@@ -1108,10 +1131,12 @@ recode_new_request (RECODE_OUTER outer)
 bool
 recode_delete_request (RECODE_REQUEST request)
 {
-  if (request->sequence_array)
-    free (request->sequence_array);
-  if (request->work_string)
-    free (request->work_string);
+  for (RECODE_STEP step = request->sequence_array;
+       step < request->sequence_array + request->sequence_length;
+       step++)
+    delete_step (step);
+  free (request->sequence_array);
+  free (request->work_string);
   free (request);
   return true;
 }
@@ -1180,9 +1205,7 @@ recode_buffer_to_buffer (RECODE_CONST_REQUEST request,
   task->output.cursor = *output_buffer_pointer;
   task->output.limit = *output_buffer_pointer + *output_allocated_pointer;
 
-  task->strategy = RECODE_SEQUENCE_IN_MEMORY;
-  success = recode_perform_task (task);
-  guarantee_nul_terminator (task);
+  success = recode_perform_task (task) && guarantee_nul_terminator (task);
   *output_buffer_pointer = task->output.buffer;
   *output_length_pointer = task->output.cursor - task->output.buffer;
   *output_allocated_pointer = task->output.limit - task->output.buffer;
@@ -1208,7 +1231,6 @@ recode_buffer_to_file (RECODE_CONST_REQUEST request,
   task->input.limit = input_buffer + input_length;
   task->output.file = output_file;
 
-  task->strategy = RECODE_SEQUENCE_IN_MEMORY;
   success = recode_perform_task (task);
 
   recode_delete_task (task);
@@ -1233,9 +1255,7 @@ recode_file_to_buffer (RECODE_CONST_REQUEST request,
   task->output.cursor = *output_buffer_pointer;
   task->output.limit = *output_buffer_pointer + *output_allocated_pointer;
 
-  task->strategy = RECODE_SEQUENCE_IN_MEMORY;
-  success = recode_perform_task (task);
-  guarantee_nul_terminator (task);
+  success = recode_perform_task (task) && guarantee_nul_terminator (task);
   *output_buffer_pointer = task->output.buffer;
   *output_length_pointer = task->output.cursor - task->output.buffer;
   *output_allocated_pointer = task->output.limit - task->output.buffer;
